#include <windows.h>
#include <winuser.h>
#include <jni.h>
#include <jni_md.h>
#include <jawt.h>
#include <jawt_md.h>
#include <assert.h>
#include "jWinAPI.h"


#include "jWinApi.struct"
#include "jWinApi.helpers"


JavaVM *jvm;

#define JWINAPI_DLL_VERSION "1.3"

HDDEDATA  StdDDECallback(UINT wType,
                                 UINT wFmt,
                                 HCONV hConv,
                                 HDDEDATA hsz1,
                                 HDDEDATA hsz2,
                                 HDDEDATA hData,
                                 HDDEDATA dwData1,
                                 HDDEDATA dwData2);



JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *the_jvm, void *reserved) {
  jvm = the_jvm;
  return JNI_VERSION_1_2;
}

JNIEnv *JNI_GetEnv() {
  JNIEnv *env;
  jvm->GetEnv((void **)&env, JNI_VERSION_1_2);
  return env;
}

/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    GetCurrentThreadId
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_com_devexperts_ddelink_WinAPI_GetCurrentThreadId
  (JNIEnv *env, jclass winapi) {
   return (jint) GetCurrentThreadId();
}

/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    GetMessage
 * Signature: (Lcom/devexperts/ddelink/winapi/MSG;JJ)I
 */
JNIEXPORT jint JNICALL Java_com_devexperts_ddelink_WinAPI_GetMessage
  (JNIEnv *env, jclass winapi, jobject jmsg, jlong filter_from, jlong filter_to) {

    MSG msg;
    BOOL result = GetMessage(&msg, NULL, filter_from, filter_to);
    if (env->ExceptionCheck()) {
    return FALSE;
    }
    StructToClass_MSG(env, jmsg, &msg);
    return result;
}



/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    PeekMessage
 * Signature: (Lcom/devexperts/ddelink/winapi/MSG;JJ)I
 */
JNIEXPORT jint JNICALL Java_com_devexperts_ddelink_WinAPI_PeekMessage
  (JNIEnv *env, jclass winapi, jobject jmsg, jlong filter_from, jlong filter_to, jlong remove_message) {
    MSG msg;
    BOOL result = PeekMessage(&msg, NULL, filter_from, filter_to, remove_message);
    if (env->ExceptionCheck()) {
    return FALSE;
    }
    StructToClass_MSG(env, jmsg, &msg);
    return result;
}


/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    PostThreadMessage
 * Signature: (IIIJ)I
 */
JNIEXPORT jint JNICALL Java_com_devexperts_ddelink_WinAPI_PostThreadMessage
  (JNIEnv *env, jclass winapi, jint thread_id, jint msg, jint wparam, jlong lparam) {
    jint result = PostThreadMessage(thread_id, msg, wparam, lparam);
    return result;
}

/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DispatchMessage
 * Signature: (Lcom/devexperts/ddelink/winapi/MSG;)J
 */
JNIEXPORT jlong JNICALL Java_com_devexperts_ddelink_WinAPI_DispatchMessage
  (JNIEnv *env, jclass winapi, jobject jmsg) {
  MSG msg;
  ClassToStruct_MSG(env, jmsg, &msg);
  long result = DispatchMessage(&msg);
  return result;
}

JNIEXPORT jint JNICALL Java_com_devexperts_ddelink_WinAPI_DdeInitialize
  (JNIEnv *env, jclass winapi, jint afCmd, jint ulRes) {
  DWORD dde_inst = 0;
  jint result = DdeInitialize(&dde_inst, (PFNCALLBACK)&StdDDECallback, afCmd, ulRes);
  jclass dde_class = env->FindClass("com/devexperts/ddelink/DDEManager");
  if (dde_class == NULL) {
    return NULL; /* class not found */
  }

  jmethodID reg_method = env->GetStaticMethodID(dde_class, "DDERegistred", "(IJ)V");
  if (reg_method == NULL) {
    return NULL;  /* method not found */
  }
  env->CallStaticVoidMethod(dde_class, reg_method,
                (jint)GetCurrentThreadId(),
                            (jint)dde_inst);
  return result;
}

/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DdeUninitialize
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_com_devexperts_ddelink_WinAPI_DdeUninitialize
  (JNIEnv *env, jclass winapi, jlong idInst) {
  return DdeUninitialize(idInst) ? JNI_TRUE : JNI_FALSE;
}

/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DdeCreateStringHandle
 * Signature: (JLjava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_com_devexperts_ddelink_WinAPI_DdeCreateStringHandle
  (JNIEnv *env, jclass winapi, jlong idInst, jstring psz) {

  const char *str = env->GetStringUTFChars(psz, NULL);
  jint result = (jint)DdeCreateStringHandle (idInst, str, CP_WINANSI);
  env->ReleaseStringUTFChars(psz, str);
  return result;
}

/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DdeFreeStringHandle
 * Signature: (JI)Z
 */
JNIEXPORT jboolean JNICALL Java_com_devexperts_ddelink_WinAPI_DdeFreeStringHandle
  (JNIEnv *env, jclass winapi, jlong idInst, jint hsz) {
  return DdeFreeStringHandle(idInst, (HSZ)hsz) ? JNI_TRUE : JNI_FALSE;
}



/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DdeQueryString
 * Signature: (JI)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_com_devexperts_ddelink_WinAPI_DdeQueryString
  (JNIEnv *env, jclass winapi, jlong idInst, jint hsz) {
  DWORD length = DdeQueryString(idInst, (HSZ)hsz, NULL, 0, CP_WINANSI);
  char * buf = (char *)malloc(length + 1);
  DdeQueryString(idInst, (HSZ)hsz, buf, length + 1, CP_WINANSI);
  jstring result = env->NewStringUTF(buf);
  free(buf);
  return result;
}

/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DdeCmpStringHandles
 * Signature: (III)Z
 */
JNIEXPORT jboolean JNICALL Java_com_devexperts_ddelink_WinAPI_DdeCmpStringHandles
  (JNIEnv *env, jclass winapi, jint hsz1, jint hsz2) {
  return (0 == DdeCmpStringHandles((HSZ)hsz1, (HSZ)hsz2))  ? JNI_TRUE : JNI_FALSE;
}

/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DdeGetDataAsString
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_com_devexperts_ddelink_WinAPI_DdeGetDataAsString
  (JNIEnv *env , jclass winapi, jint hData) {
  char *buf = (char*)DdeAccessData((HDDEDATA)hData, NULL);
  jstring result = env->NewStringUTF(buf);
  DdeUnaccessData((HDDEDATA)hData);
  return result;
}


/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DdeNameService
 * Signature: (JIII)I
 */
JNIEXPORT jint JNICALL Java_com_devexperts_ddelink_WinAPI_DdeNameService
  (JNIEnv *env, jclass winapi, jlong idInst, jint hsz1, jint hsz2, jint afCmd) {
  return (jint)DdeNameService(idInst, (HSZ)hsz1, (HSZ)hsz2, afCmd);
}

/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DdeGetLastError
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_com_devexperts_ddelink_WinAPI_DdeGetLastError
  (JNIEnv *env, jclass winapi, jlong idInst) {
  return (jint)DdeGetLastError(idInst);
}

/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DdePostAdvise
 * Signature: (JII)Z
 */
JNIEXPORT jboolean JNICALL Java_com_devexperts_ddelink_WinAPI_DdePostAdvise
  (JNIEnv *env, jclass winapi, jlong idInst, jint hsz1, jint hsz2) {
  return DdePostAdvise(idInst, (HSZ)hsz1, (HSZ)hsz2);
}


/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    DdeCreateDataHandle
 * Signature: (J[BII)I
 */
JNIEXPORT jint JNICALL Java_com_devexperts_ddelink_WinAPI_DdeCreateDataHandle
  (JNIEnv *env, jclass winapi, jlong idInst, jbyteArray data, jint hszItem, jint wFmt) {
  int data_size = env->GetArrayLength(data);
  jbyte* data_bytes = env->GetByteArrayElements(data, NULL);
  if (NULL == data_bytes) {
    return 0;
  }
  jint result = (jint)DdeCreateDataHandle(idInst, (byte *)data_bytes, data_size, 0, (HSZ)hszItem, wFmt,0);
  env->ReleaseByteArrayElements(data, data_bytes, JNI_ABORT);
  return result;
}




/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    getLibraryVersion
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_com_devexperts_ddelink_WinAPI_getLibraryVersion
  (JNIEnv *env, jclass winapi) {
  char *buf = JWINAPI_DLL_VERSION;
  jstring result = env->NewStringUTF(buf);
  return result;
}


/*
 * Class:     com_devexperts_ddelink_WinAPI
 * Method:    getAvailablePhysicalMemory
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_com_devexperts_ddelink_WinAPI_getAvailablePhysicalMemory
  (JNIEnv *env, jclass winapi) {
    MEMORYSTATUS stat;
    GlobalMemoryStatus (&stat);
    return stat.dwTotalPhys;
}


HDDEDATA  StdDDECallback(UINT uType,
                                 UINT uFmt,
                                 HCONV hConv,
                                 HDDEDATA hsz1,
                                 HDDEDATA hsz2,
                                 HDDEDATA hData,
                                 HDDEDATA dwData1,
                                 HDDEDATA dwData2) {
  JNIEnv *env = JNI_GetEnv();


  jclass dde_class = env->FindClass("com/devexperts/ddelink/DDEManager");
  if (dde_class == NULL) {
    return NULL; /* class not found */
  }
 
  jmethodID callback = env->GetStaticMethodID(dde_class, "DDECallback", "(IIIIIIIII)I");
  if (callback == NULL) {
    return NULL;  /* method not found */
  }
  jint result = env->CallStaticIntMethod(dde_class, callback, 
        (jint)GetCurrentThreadId(),
        (jint)uType,          
                (jint)uFmt,   
                (jint)hConv, 
                (jint)hsz1,    
                (jint)hsz2,    
                (jint)hData,
                (jint)dwData1,
                (jint)dwData2);
  return (HDDEDATA)result;
}


